# coding: utf-8
from functools import singledispatch


def prepare_query_params(**kwargs):
    """
    Prepares given parameters to be used in querystring.
    """
    return [
        (sub_key, sub_value)
        for key, value in kwargs.items()
        for sub_key, sub_value in expand(value, key)
        if sub_value is not None
    ]


class Filter(dict):
    """
    Custom subtype to apply special behaviour.
    """

    def __init__(self, connector='AND', **kwargs):
        self.connector = connector
        super().__init__(**kwargs)


class Sort(list):

    def __init__(self, args):
        if not isinstance(args, (list, tuple, set)):
            args = [args]
        super().__init__(args)


OPERATORS = {
    'ne': 'NOT_EQUAL_TO',
    'lt': 'LESS_THAN',
    'lte': 'LESS_THAN_OR_EQUAL_TO',
    'gt': 'GREATER_THAN',
    'gte': 'GREATER_THAN_OR_EQUAL_TO',
    'like': 'LIKE',
    'not_like': 'NOT_LIKE',
    'null': 'NULL',
    'not_null': 'NOT_NULL',
    'true': 'TRUE',
    'false': 'FALSE',
}


@singledispatch
def expand(value, key):
    yield key, value


@expand.register(dict)
def expand_dict(value, key):
    for dict_key, dict_value in value.items():
        yield '%s[%s]' % (key, dict_key), dict_value


@expand.register(Filter)
def expand_filter(value, key):
    for dict_key, dict_value in value.items():
        if isinstance(dict_value, (list, tuple, set)):
            for sub_value in dict_value:
                yield '%s[%s][]' % (key, dict_key), sub_value
        else:
            try:
                field_name, operator = dict_key.split('__')
                param_name = '%s[%s][%s]' % (key, field_name, OPERATORS[operator])
            except ValueError:
                if value.connector == 'OR':
                    param_name = '%s[%s][%s]' % (key, value.connector, dict_key)
                else:
                    param_name = '%s[%s]' % (key, dict_key)
            yield param_name, dict_value


@expand.register(Sort)
def expand_sort(value, key):
    for field_name in value:
        order = 'asc'
        if field_name.startswith('-'):
            order = 'desc'
            field_name = field_name[1:]
        yield '%s[%s]' % (key, field_name), order


@expand.register(list)
@expand.register(tuple)
@expand.register(set)
def expand_lists(value, key):
    for inner_value in value:
        yield '%s[]' % key, inner_value



# coding: utf-8
from .advertiser import Advertiser, AdvertiserManager
from .conversion import Conversion, ConversionManager
from .core import ModelManager
from .country import Country, CountryManager
from .goal import Goal, GoalManager
from .offer import Offer, OfferManager


MODEL_MANAGERS = (
    AdvertiserManager, ConversionManager, CountryManager, GoalManager, OfferManager
)



# coding: utf-8
from .core import Model, ModelManager


class Goal(Model):
    """
    A Goal for an Offer.
    """
    generic_methods = ('update',)


class GoalManager(ModelManager):
    model = Goal
    name = 'goals'
    generic_methods = (
        'create',
        'update',
        'find_by_id',
        'find_all',
    )



# coding: utf-8
from collections import defaultdict

from ..utils import Filter, Sort


class SelectiveInheritanceMeta(type):
    """
    Allows to inherit only methods from `generic_methods` attribute.
    """
    generic_methods = defaultdict(dict)

    def __new__(mcs, name, bases, members):
        if not bases:
            for key, value in list(members.items()):
                if getattr(value, '_is_generic', None):
                    mcs.generic_methods[name][key] = members.pop(key)
        else:
            for method in members.get('generic_methods', ()):
                members[method] = mcs.generic_methods[bases[0].__name__][method]
        return super().__new__(mcs, name, bases, members)


def generic_method(method):
    """
    Marks method as generic.
    """
    method._is_generic = True
    return method


class Model(metaclass=SelectiveInheritanceMeta):
    """
    Abstract model for HasOffers entity.
    """
    generic_methods = ()

    def __init__(self, manager, **kwargs):
        self._manager = manager
        self._data = kwargs

    def __str__(self):
        return '%s: %s' % (self.__class__.__name__, self._data.get('id'))

    def __repr__(self):
        return '<%s>' % self

    def __getitem__(self, item):
        return self._data[item]

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self['id'] == other['id']

    @generic_method
    def update(self, **kwargs):
        return self._manager.update(self['id'], **kwargs)


class ModelManager(metaclass=SelectiveInheritanceMeta):
    """
    Proxy for API methods with predefined model.
    Used to structure API. Every manager works only with single model methods.
    """
    model = None
    name = None
    generic_methods = ()

    def __init__(self, api):
        self.api = api

    def init_instance(self, data):
        return self.model(manager=self, **data)

    def _call(self, method, **kwargs):
        return self.api._call(self.model.__name__, method, **kwargs)

    @generic_method
    def create(self, **kwargs):
        return self._call('create', data=kwargs)

    @generic_method
    def update(self, id, **kwargs):
        return self._call('update', id=id, data=kwargs)

    @generic_method
    def find_by_id(self, id, contain=None):
        return self._call('findById', id=id, contain=contain)

    @generic_method
    def find_all(self, sort=(), limit=None, page=None, contain=None, **kwargs):
        return self._call(
            'findAll',
            filters=Filter(**kwargs), sort=Sort(sort), limit=limit, page=page, contain=contain, single_result=False
        )



# coding: utf-8
from .core import Model, ModelManager


class Country(Model):
    pass


class CountryManager(ModelManager):
    model = Country
    name = 'countries'

    def init_instance(self, data):
        if 'id' not in data:
            data = list(data.values())[0]
        return super().init_instance(data)
